#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.sql
  ~~~~~~~~~~~~~

  The default :class:`ElementAdapter` that ships with Chanjo. Provides a basic
  interface to a `SQLite` database using the `SQLAlchemy` ORM. The SQL
  structure extends a parallel project `Elemental` that provides the tables
  and relationships between genes, transcripts, and exons.

  The module defines a few extra ORM objects for adding sample specific
  coverage annotations.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details
"""
import sqlalchemy as sa
from sqlalchemy.orm import relationship, backref

from elemental.core import Base, ElementalDB


# ---------------------------------------------------------
#  Data ORM classes
# ---------------------------------------------------------
class GeneData(Base):
  """
  Stores coverage metrics for a single gene and a given sample. It has a
  many-to-one relationship with it's parent gene object through it's
  ``element_id`` attribute.

  :param str element_id: HGNC gene symbol
  :param str sample_id: Unique sample identifier
  :param int group_id: "Unique" group identifier
  :param float coverage: Average coverage for the gene
  :param float completeness: Ratio of adequately covered bases
  """
  __tablename__ = "GeneData"

  id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
  coverage = sa.Column(sa.Float)
  completeness = sa.Column(sa.Float)

  # These column maps coverage/completeness to an individual+group
  sample_id = sa.Column(sa.String)
  group_id = sa.Column(sa.Integer)

  # Genetic relationship
  element_id = sa.Column(sa.String, sa.ForeignKey("Gene.id"))
  element = relationship("Gene", backref=backref("data"))

  def __init__(self, element_id, sample_id=None, group_id=None,
               coverage=None, completeness=None):
    super(GeneData, self).__init__()
    
    self.element_id = element_id
    self.sample_id = sample_id
    self.group_id = group_id
    self.coverage = coverage
    self.completeness = completeness


class TranscriptData(Base):
  """
  Stores coverage metrics for a single transcript and a given sample. It has a
  many-to-one relationship with it's parent transcript object through it's
  ``element_id`` attribute.

  :param str element_id: Unique CCDS transcript identifier
  :param str sample_id: Unique sample identifier
  :param int group_id: "Unique" group identifier
  :param float coverage: Average coverage for the transcript
  :param float completeness: Ratio of adequately covered bases
  """
  __tablename__ = "TranscriptData"

  id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
  coverage = sa.Column(sa.Float)
  completeness = sa.Column(sa.Float)

  # These column maps coverage/completeness to an individual+group
  sample_id = sa.Column(sa.String)
  group_id = sa.Column(sa.Integer)

  # Genetic relationship
  element_id = sa.Column(sa.String, sa.ForeignKey("Transcript.id"))
  element = relationship("Transcript", backref=backref("data"))

  def __init__(self, element_id, sample_id=None, group_id=None,
               coverage=None, completeness=None):
    super(TranscriptData, self).__init__()
    
    self.element_id = element_id
    self.sample_id = sample_id
    self.group_id = group_id
    self.coverage = coverage
    self.completeness = completeness


class ExonData(Base):
  """
  Stores coverage metrics for a single exon and a given sample. It has a
  many-to-one relationship with it's parent exon object through it's
  ``element_id`` attribute.

  :param str element_id: HGNC gene symbol
  :param str sample_id: Unique sample identifier
  :param int group_id: "Unique" group identifier
  :param float coverage: Average coverage for the exon
  :param float completeness: Ratio of adequately covered bases
  """
  __tablename__ = "ExonData"

  id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
  coverage = sa.Column(sa.Float)
  completeness = sa.Column(sa.Float)

  # These column maps coverage/completeness to an individual+group
  sample_id = sa.Column(sa.String)
  group_id = sa.Column(sa.Integer)

  # Genetic relationship
  element_id = sa.Column(sa.String, sa.ForeignKey("Exon.id"))
  element = relationship("Exon", backref=backref("data"))

  def __init__(self, element_id, sample_id=None, group_id=None,
               coverage=None, completeness=None):
    super(ExonData, self).__init__()
    
    self.element_id = element_id
    self.sample_id = sample_id
    self.group_id = group_id
    self.coverage = coverage
    self.completeness = completeness


# =========================================================
#   The actual Element Adapter
# ---------------------------------------------------------
class ElementAdapter(ElementalDB):
  """
  SQLAlchemy-based :class:`ElementAdapter` for Chanjo. Inherits the basic SQL
  table structure from `ElementalDB`.

  .. note::

    For testing pourposes; use ":memory:" as the ``path`` parameter to set up
    an in-memory instance of the database.

  :param str path: Path to the database to connect to
  :param bool debug: Whether to print logging information (optional)
  """
  def __init__(self, path, debug=False):
    super(ElementAdapter, self).__init__(path, debug=debug)

    # Add new data classes to the supported ORM classes
    self.classes.update({
      "gene_data": GeneData,
      "transcript_data": TranscriptData,
      "exon_data": ExonData
    })

  def transcriptStats(self, sample_id):
    """
    <public> Calculates transcript level metrics to annotate transcripts.
    Requires all related exons to already be properly annotated.

    What's happening is that we are summing read depths and adequately covered
    bases for each exon in a transcript and then dividing those numbers by the
    total exon length of the transcript.

    .. note::

      Transcript annotation needs to be carried out before annotating genes!

    :param str sample_id: Sample ID to match with coverage annotations
    :returns: List of tuples: ``(<tx_id>, <coverage>, <completeness>)``
    """
    # I might turn this into a proper SQLAlchemy verison but for now.
    rawSQL = """
    SELECT Exon_Transcript.transcript_id,
           sum((Exon.end-Exon.start+1) * ExonData.coverage) / sum(Exon.end-Exon.start+1) AS coverage,
           sum((Exon.end-Exon.start+1) * ExonData.completeness) / sum(Exon.end-Exon.start+1) AS completeness
    FROM Exon
    INNER JOIN Exon_Transcript ON Exon.id=Exon_Transcript.exon_id
    INNER JOIN ExonData ON Exon.id=ExonData.element_id
    WHERE ExonData.sample_id="{sample_id}"
    GROUP BY Exon_Transcript.transcript_id
    """.format(sample_id=sample_id)

    return self.session.execute(rawSQL).fetchall()

  def geneStats(self, sample_id):
    """
    <public> Calculates gene level metrics to annotate genes. Requires all
    related transcripts to already be properly annotated.

    What's happening is that we are simply taking the average of the metrics
    on the transcript level and applying that as gene metrics. This gives a
    decent, albeit not perfect, represenation of gene level metrics.

    .. note::

      Annotation of transcripts needs to be acomplished before annotating genes!

    :param str sample_id: Sample ID to match with coverage annotations
    :returns: List of tuples: ``(<gene_id>, <coverage>, <completeness>)``
    """
    rawSQL = """
    SELECT Gene.id,
           avg(TranscriptData.coverage),
           avg(TranscriptData.completeness)
    FROM Transcript
    INNER JOIN Gene ON Transcript.gene_id=Gene.id
    INNER JOIN TranscriptData ON Transcript.id=TranscriptData.element_id
    WHERE TranscriptData.sample_id="{sample_id}"
    GROUP BY Transcript.gene_id
    """.format(sample_id=sample_id)

    return self.session.execute(rawSQL).fetchall()



#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.bam
  ~~~~~~~~~~~~~

  The default :class:`CoverageAdapter` that ships with Chanjo. Talks directly
  to a BAM alignment file to extract read depth data.

  Depends on the Pysam_ package and requires Samtools_ to be installed in your
  ``$PATH`` to work.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details

  .. _Pysam: http://www.cgat.org/~andreas/documentation/pysam/contents.html
  .. _Samtools: http://samtools.sourceforge.net/
"""
import pysam
import numpy as np


class CoverageAdapter(pysam.Samfile):
  """
  Adapter for interfacing directly with BAM alignment files. Inherits from
  :class:`Samfile` which requires a second init parameter that tells it to
  expect a *binary* BAM-file rather than the plain text cousin SAM-file format.

  .. code-block:: python

    >>> from chanjo.bam import CoverageAdapter
    >>> bam_path = "/path/to/bam/file.bam"
    >>> adapter = CoverageAdapter(bam_path)

  :param str bam_path: Path to a BAM alignment file
  """

  def __init__(self, bam_path):
    super(CoverageAdapter, self).__init__(bam_path, "rb")

  def read(self, chrom, start, end):
    """
    <public> Generates a list of read depths for each position between
    (start, end). The `numpy` array is used to optimize performance when
    building and slicing the list.

    This method depends on `Pysam` >=0.7.5 since using the `truncate` which
    wasn't working in previous versions.

    .. code-block:: python

      >>> adapter.read("17", 0, 5)
      array([3., 3., 4., 4., 5., 4.])

    .. note::

      Positions are 0,0-based throughout `Chanjo`. If start=0, end=9 you should
      expect the 10 read depths for position 1-10 to be returned.

    :param str chrom: The chromosome ID (str) of interest
    :param int start: The first position of the interval (0-based)
    :param int end: The last position of the interval (0-based)
    :returns: A numpy array of read depths for *each* position in the interval
    """
    # Generate a list of 0 read depth for each position
    positions = np.zeros(end+1-start)

    # Start Pileup iterator and walk through each position in the interval
    # `truncate` will make sure it starts and ends on the given positions!
    # +1 to end because pysam otherwise stops one base short by default
    for col in self.pileup(str(chrom), start, end+1, truncate=True):

      # Overwrite the read depth in the correct position
      # This will allow simple slicing to get at the positions of interest
      positions[col.pos - start] = col.n

    return positions



#!/usr/bin/env python
# coding: utf-8
"""
  chanjo.core
  ~~~~~~~~~~~~~

  This module implements the glue and controlling hub of `Chanjo`. The
  :class:`Hub` is also where the adapters are plugged in to. This will likely
  be pretty much the only part of the package you directly interact with if
  you decide to use the Pythonic API.

  :copyright: (c) 2013 by Robin Andeer
  :license: MIT, see LICENSE for more details
"""


class Hub(object):
  """
  The :class:`Hub` is the core component of `Chanjo`. It's able to prepare
  data and annotate elements, calculate coverage for a range of genomic
  positions, and interact with the element data store.

  .. code-block:: python

    from chanjo.core import Hub
    from chanjo.bam import CoverageAdapter
    from chanjo.sqlite import ElementAdapter

    bam = CoverageAdapter("/path/to/file.bam")
    sql = ElementAdapter("/path/to/sqlite.db")
    hub = Hub(bam, sql)

  :param str coverageAdapter: (optional) Plug in the adapter during init
  :param str elementAdapter: (optional) Plug in the adapter during init

  """
  def __init__(self, coverageAdapter=None, elementAdapter=None):
    super(Hub, self).__init__()

    # Set up the adapters
    if coverageAdapter and elementAdapter:
      self.connect(coverageAdapter, elementAdapter)

  def connect(self, coverageAdapter, elementAdapter):
    """
    <public> Plugs in the required adapters.

    .. code-block:: python

      >>> bam_path = "/path/to/file.bam"
      >>> cov_path = "/path/to/sqlite.db"
      >>> hub.connect(CoverageAdapter(bam_path), ElementAdapter(cov_path))

    :param str coverageAdapter: An instance of a :class:`CoverageAdapter`
    :param str elementAdapter:  An instance of a :class:`ElementAdapter`
    """
    # Customizable adapters
    self.cov = coverageAdapter
    self.db = elementAdapter

  def annotate(self, element, cutoff=10, splice=False):
    """
    <public> Annotates each related exon with coverage data.

    .. code-block:: python

      >>> genes = hub.db.get("gene", ["GIT1", "EGFR", "BRCA1"])
      >>> for gene in genes:
      >>>   hub.annotate(gene, 15)

    :param object element: One element object (gene/transcript)
    :param int cutoff: (optional) Min read depth for completeness [default: 10]
    :param bool splice: (optional) Include splice sites for each exon (+/- 2)
    """
    start = element.start
    end = element.end

    # Include splice sites (+/- 2 bases) when reading from BAM-file
    if splice:
      start -= 2
      end += 2

    # Both transcripts and genes can be used to select exons to annotate
    depth = self.cov.read(element.chrom, start, end)

    # Preallocate list for each exon of the element
    exons = [None]*len(element.exons)

    # Get the exons related to the element
    for i, exon in enumerate(element.exons):
      ex_start = exon.start
      ex_end = exon.end

      # Include splice sites for the exon
      if splice:
        ex_start -= 2
        ex_end += 2

      # Relative start and end positions to slice the ``depth`` array
      rel_start = ex_start - start
      rel_end = ex_end - start

      # Do the heavy lifting
      # +1 to end because ``end`` is 0-based and slicing is 0,1-based
      (coverage,
       completeness) = self.calculate(depth[rel_start:rel_end+1], cutoff)

      exons[i] = {
        "element_id": exon.id,
        "coverage": coverage,
        "completeness": completeness
      }

    return exons

  def calculate(self, depths, cutoff):
    """
    <public> Calculates both coverage and completeness for an interval.

    .. code-block:: python

      >>> gene = hub.db.get("gene", "C3")
      >>> hub.coverage(gene.chrom, gene.intervals, 15)
      (13.43522398231, 0.434122133123)

    :param list depths: List/array of the read depth for each position/base
    :param int cutoff: The cutoff for lowest passable read depth (completeness)
    :returns: ``(<coverage (float)>, <completeness (float)>)``
    :rtype: tuple
    """
    # Initialize
    totBaseCount = float(len(depths))
    readCount = 0
    passedCount = 0

    for depth in depths:
      # Add the number of overlapping reads
      readCount += depth

      # Add the position if it passes `cutoff`
      if depth >= cutoff:
        passedCount += 1

    # totBaseCount should never be able to be 0! Exons be >= 1 bp long
    return readCount / totBaseCount, passedCount / totBaseCount



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Bootstrap distribute unless already installed
from distribute_setup import use_setuptools
use_setuptools()

import os
import sys

from setuptools import setup, find_packages
import chanjo

# Shortcut for publishing to Pypi
# Source: https://github.com/kennethreitz/tablib/blob/develop/setup.py
if sys.argv[-1] == "publish":
  os.system("python setup.py sdist upload")
  sys.exit()

setup(
  name="chanjo",
  version=chanjo.__version__,
  packages=find_packages(exclude=["tests"]),
  scripts=[
    "scripts/chanjo"
  ],

  # Project dependencies
  install_requires = [
    "pysam",
    "numpy",
    "sqlalchemy",
    "docopt",
    "elemental",
    "path.py"
  ],

  # Packages required for testing
  tests_require = [
    "nose"
  ],

  # Metadate for upload to Pypi
  author="Robin Andeer",
  author_email="robin.andeer@gmail.com",
  description="A coverage analysis package for clinical sequencing.",
  long_description = (open('README.rst').read()),
  license="MIT",
  keywords="coverage sequencing clinical exome",
  platform="UNIX",
  url="http://chanjo.readthedocs.org/",
  download_url="https://github.com/robinandeer/chanjo",
  classifiers=[
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Programming Language :: Python :: 2.7",
    "Topic :: Scientific/Engineering :: Bio-Informatics"
  ]
)